package de.hft.stuttgart.citydoctor2.gui;

import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.utils.Localization;
import javafx.application.Platform;
import javafx.embed.swing.SwingFXUtils;
import javafx.event.Event;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.SnapshotParameters;
import javafx.scene.chart.BarChart;
import javafx.scene.chart.CategoryAxis;
import javafx.scene.chart.NumberAxis;
import javafx.scene.chart.XYChart;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.image.Image;
import javafx.scene.image.WritableImage;
import javafx.scene.layout.VBox;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.util.StringConverter;

public class WriteReportDialog {

	private static Logger logger = LogManager.getLogger(WriteReportDialog.class);

	private Stage stage;
	private CityDoctorController controller;
	private MainWindow mainWindow;

	@FXML
	private CheckBox pdfCheckBox;

	@FXML
	private TextField pdfFileField;

	@FXML
	private Button selectPdfBtn;

	@FXML
	private CheckBox xmlCheckBox;

	@FXML
	private TextField xmlFileField;

	@FXML
	private Button selectXmlFile;

	@FXML
	private Button cancelBtn;

	@FXML
	private Button writeBtn;

	@FXML
	private BarChart<String, Number> barChart;

	@FXML
	private Button saveImageBtn;

	@FXML
	private Label errorStatisticsLabel;
	
	public WriteReportDialog(Stage parent, CityDoctorController controller, MainWindow mainWindow) throws IOException {
		FXMLLoader loader = new FXMLLoader(WriteReportDialog.class.getResource("WriteReportDialog.fxml"));
		loader.setController(this);
		VBox box = loader.load();
		this.controller = controller;
		this.mainWindow = mainWindow;
		stage = new Stage();
		stage.getIcons().add(new Image(MainWindow.class.getResourceAsStream("icons/CityDoctor-Logo-rot_klein.jpg")));
		stage.setScene(new Scene(box));
		stage.initOwner(parent);
		stage.initModality(Modality.APPLICATION_MODAL);
		stage.setTitle("Write Reports");
		stage.setOnHidden(we -> barChart.getData().clear());
		setMaximumWidthOfBarChartColumns();
	}

	private void setMaximumWidthOfBarChartColumns() {
		double maxBarWidth = 40;
		double minCategoryGap = 10;

		// setting maximum width of columns in bar chart
		// from https://stackoverflow.com/questions/27302875/set-bar-chart-column-width-size
		stage.getScene().widthProperty().addListener((obs, n, n1) -> {
			if (barChart.getData().isEmpty())
				return;

			setCategoryWidthWhenWindowIsBigger(maxBarWidth, n, n1);
			if (n != null && (n1.doubleValue() < n.doubleValue()) && barChart.getCategoryGap() > minCategoryGap) {
				double barWidth = 0;
				CategoryAxis xAxis = (CategoryAxis) barChart.getXAxis();
				do {
					double catSpace = xAxis.getCategorySpacing();
					double avilableBarSpace = catSpace - (minCategoryGap + barChart.getBarGap());
					barWidth = Math.min(maxBarWidth,
							(avilableBarSpace / barChart.getData().size()) - barChart.getBarGap());
					avilableBarSpace = (barWidth + barChart.getBarGap()) * barChart.getData().size();
					barChart.setCategoryGap(catSpace - avilableBarSpace - barChart.getBarGap());
				} while (barWidth < maxBarWidth && barChart.getCategoryGap() > minCategoryGap);
			}
		});
	}

	private void setCategoryWidthWhenWindowIsBigger(double maxBarWidth, Number n, Number n1) {
		if (n != null && (n1.doubleValue() > n.doubleValue())) {
			double barWidth = 0;
			CategoryAxis xAxis = (CategoryAxis) barChart.getXAxis();
			do {
				double catSpace = xAxis.getCategorySpacing();
				double avilableBarSpace = catSpace - (barChart.getCategoryGap() + barChart.getBarGap());
				barWidth = (avilableBarSpace / barChart.getData().size()) - barChart.getBarGap();
				if (barWidth > maxBarWidth) {
					avilableBarSpace = (maxBarWidth + barChart.getBarGap()) * barChart.getData().size();
					barChart.setCategoryGap(catSpace - avilableBarSpace - barChart.getBarGap());
				}
			} while (barWidth > maxBarWidth);
		}
	}

	public void initialize() {
		writeBtn.setText(Localization.getText("WriteReportDialog.writeBtn"));
		cancelBtn.setText(Localization.getText("WriteReportDialog.cancelBtn"));
		errorStatisticsLabel.setText(Localization.getText("WriteReportDialog.errorStatisticsLabel"));
		saveImageBtn.setText(Localization.getText("WriteReportDialog.saveImageBtn"));
		selectPdfBtn.setText(Localization.getText("WriteReportDialog.selectPdfBtn"));
		selectXmlFile.setText(Localization.getText("WriteReportDialog.selectXmlFile"));
		barChart.getXAxis().setLabel(Localization.getText("WriteReportDialog.xAxisLabel"));
		barChart.getYAxis().setLabel(Localization.getText("WriteReportDialog.yAxisLabel"));
		
		// formatter to display only whole numbers
		NumberAxis yAxis = (NumberAxis) barChart.getYAxis();
		yAxis.setTickLabelFormatter(new StringConverter<Number>() {

			@Override
			public String toString(Number object) {
				if (object.doubleValue() % 1 < 0.00000001) {
					return String.valueOf(object.intValue());
				} else {
					return "";
				}
			}

			@Override
			public Number fromString(String string) {
				return Integer.parseInt(string);
			}
		});
		saveImageBtn.setOnAction(ae -> {
			FileChooser chooser = new FileChooser();
			chooser.setInitialDirectory(new File("."));
			chooser.getExtensionFilters().add(new ExtensionFilter("PNG-Image", "*.png"));
			String fName = controller.getFileName();
			fName = fName.substring(0, fName.lastIndexOf('.'));
			chooser.setInitialFileName(fName + "_errors.png");
			File imageFile = chooser.showSaveDialog(stage);
			if (imageFile != null) {
				WritableImage snapshot = barChart.snapshot(new SnapshotParameters(), null);
				try {
					ImageIO.write(SwingFXUtils.fromFXImage(snapshot, null), "png", imageFile);
				} catch (IOException e) {
					logger.error("Failed to save image", e);
					mainWindow.showExceptionDialog(e);
				}
			}
		});

		pdfCheckBox.selectedProperty().addListener((obs, oldV, newV) -> {
			selectPdfBtn.setDisable(!newV);
			pdfFileField.setDisable(!newV);
		});

		xmlCheckBox.selectedProperty().addListener((obs, oldV, newV) -> {
			selectXmlFile.setDisable(!newV);
			xmlFileField.setDisable(!newV);
		});

		selectPdfBtn.setOnAction(ae -> {
			FileChooser chooser = new FileChooser();
			chooser.setTitle("Select PDF File");
			chooser.getExtensionFilters().add(new ExtensionFilter("PDF-Files", ".pdf"));
			File pdfFile = chooser.showSaveDialog(stage);
			if (pdfFile != null) {
				pdfFileField.setText(pdfFile.getAbsolutePath());
			}
		});

		selectXmlFile.setOnAction(ae -> {
			FileChooser chooser = new FileChooser();
			chooser.setTitle("Select XML File");
			chooser.getExtensionFilters().add(new ExtensionFilter("XML-Files", ".xml"));
			File xmlFile = chooser.showSaveDialog(stage);
			if (xmlFile != null) {
				xmlFileField.setText(xmlFile.getAbsolutePath());
			}
		});

		initWriteButton();

		cancelBtn.setOnAction(ae -> stage.close());
	}

	private void initWriteButton() {
		writeBtn.setOnAction(ae -> {
			stage.setOnCloseRequest(Event::consume);
			cancelBtn.setDisable(true);
			writeBtn.setDisable(true);
			Thread t = new Thread(() -> {
				try {
					if (pdfCheckBox.isSelected()) {
						String file = pdfFileField.getText();
						File pdfFile = new File(file);
						File parentFile = pdfFile.getParentFile();
						if (parentFile != null) {
							parentFile.mkdirs();
						}
						controller.writePdfReport(pdfFile);
					}
					if (xmlCheckBox.isSelected()) {
						String file = xmlFileField.getText();
						File xmlFile = new File(file);
						File parentFile = xmlFile.getParentFile();
						if (parentFile != null) {
							parentFile.mkdirs();
						}
						controller.writeXmlReport(xmlFile);
					}
				} finally {
					Platform.runLater(() -> {
						stage.setOnCloseRequest(null);
						cancelBtn.setDisable(false);
						writeBtn.setDisable(false);
						stage.close();
					});
				}
			});
			t.start();
		});
	}

	public void show() {
		XYChart.Series<String, Number> series = controller.createErrorSeries();
		barChart.getData().clear();
		barChart.getData().add(series);
		stage.show();
	}

}
